from dataclasses import dataclass
from typing import Sequence
import string
import contextlib
import mmap
import os
import threading

from xkbcommon import xkb
from evdev import ecodes as e, util

from plover.oslayer.linux.wayland_connection import (
    WaylandConnection,
    wayland_keymap_event_loop,
)


@dataclass
class KeyCodeInfo:
    keycode: int
    # Other keycodes that must be held down with the keycode to send this key
    modifiers: Sequence[int] = ()


# Difference between xkbcommon keycodes and Linux EV keycodes.
# Subtract this value from xkbcommon keycodes to get Linux EV keycodes.
XKB_TO_EV_KEYCODE_OFFSET = 8

VALID_EV_KEYCODES: set[int] = set(util.find_ecodes_by_regex(r"KEY_.*")[1])

WAYLAND_AUTO_LAYOUT_NAME = "wayland-auto"

GET_WAYLAND_KEYMAP_TIMEOUT_SECONDS = 5

# Additional aliases for xkbcommon keysyms to match names in Plover dictionaries.
# Keys beginning with "XF86" are handled as a special case during xkbcommon keymap processing.
# For each xkbcommon keysym, the lowercase of the symbol name is already added to the keymap by the code.
XKB_KEY_NAME_TO_ALIASES: dict[str, list[str]] = {
    "Return": ["\n"],
    "Control_L": ["ctrl", "ctrl_l"],
    "Shift_L": ["shift"],
    "Super_L": ["super", "windows", "command"],
    "Alt_L": ["alt", "option"],
    "Tab": ["\t"],
    "Next": ["page_down"],
    "Prior": ["page_up"],
    "KP_Home": ["kp_7"],
    "KP_Up": ["kp_8"],
    "KP_Prior": ["kp_9"],
    "KP_Left": ["kp_4"],
    "KP_Begin": ["kp_5"],
    "KP_Right": ["kp_6"],
    "KP_End": ["kp_1"],
    "KP_Down": ["kp_2"],
    "KP_Next": ["kp_3"],
    "KP_Insert": ["kp_0"],
    "KP_Delete": ["kp_dot", "kp_decimal"],
}


def xkb_keycode_to_ev_keycode(keycode: int) -> int:
    return keycode - XKB_TO_EV_KEYCODE_OFFSET


def ev_keycode_to_xkb_keycode(keycode: int) -> int:
    return keycode + XKB_TO_EV_KEYCODE_OFFSET


def get_modifier_keycodes(keymap: xkb.Keymap) -> list[list[int]]:
    """
    Returns a list of xkbcommon keycodes for each non-latched and non-locked modifier
    in order of the modifier's index.
    `result[i]` is the list of keycodes for the modifier with index `i`.

    Don't consider latched or locked modifiers (e.g. NumLock)
    because Plover doesn't need to handle those for key combos.
    """
    num_mods = keymap.num_mods()
    modifier_index_to_keycodes: list[list[int]] = [[] for _ in range(num_mods)]

    for keycode in keymap:
        if xkb_keycode_to_ev_keycode(keycode) not in VALID_EV_KEYCODES:
            # Ignore keys that can't be sent or received by evdev
            continue
        # Simulate pressing the key
        keyboard_state = xkb.KeyboardState(keymap)
        key_state = keyboard_state.update_key(keycode, xkb.KeyDirection.XKB_KEY_DOWN)
        # Check if pressing the key depresses a modifier
        is_key_mod = (key_state & xkb.StateComponent.XKB_STATE_MODS_DEPRESSED) and not (
            (key_state & xkb.StateComponent.XKB_STATE_MODS_LOCKED)
            or (key_state & xkb.StateComponent.XKB_STATE_MODS_LATCHED)
        )
        if not is_key_mod:
            continue

        num_layouts = keymap.num_layouts_for_key(keycode)

        for layout in range(0, num_layouts):
            layout_is_active = keyboard_state.layout_index_is_active(
                layout, xkb.StateComponent.XKB_STATE_LAYOUT_EFFECTIVE
            )

            if not layout_is_active:
                continue

            for mod_index in range(num_mods):
                is_mod_active = keyboard_state.mod_index_is_active(
                    mod_index, xkb.StateComponent.XKB_STATE_MODS_DEPRESSED
                )
                if not is_mod_active:
                    continue

                modifier_index_to_keycodes[mod_index].append(keycode)
            # Only consider the first active layout
            break

    return modifier_index_to_keycodes


@contextlib.contextmanager
def fd_context(fd: int):
    try:
        yield fd
    finally:
        os.close(fd)


def get_wayland_keymap(timeout: float) -> xkb.Keymap:
    """Get the current keymap from the default Wayland server"""
    with WaylandConnection() as connection:
        done = False

        def timeout_thread_function():
            import time

            time.sleep(timeout)
            if not done:
                connection.shutdown()

        timeout_thread = threading.Thread(target=timeout_thread_function)
        timeout_thread.start()

        try:
            keymap_fd, keymap_size = wayland_keymap_event_loop(connection)
            done = True
        except InterruptedError:
            raise TimeoutError("Timeout retrieving keymap from Wayland")
    with (
        fd_context(keymap_fd) as keymap_fd,
        mmap.mmap(
            keymap_fd, keymap_size, flags=mmap.MAP_PRIVATE, prot=mmap.PROT_READ
        ) as keymap_file,
    ):
        xkb_context = xkb.Context()
        return xkb_context.keymap_new_from_file(keymap_file)


def generate_plover_keymap_from_xkb_keymap_and_modifiers(
    keymap: xkb.Keymap, modifier_index_to_xkb_keycode: list[list[int]]
) -> dict[str, KeyCodeInfo]:
    """
    Generate a mapping of Plover key names (key names used in Plover dictionary entries) to `KeyCodeInfo`.
    `modifier_index_to_keycode` should be the result of `get_modifier_keycodes`.
    This is a parameter to avoid recomputing the modifier keycodes if they are needed multiple times.
    """
    plover_key_to_keycode: dict[str, KeyCodeInfo] = {}

    layout_index = 0
    for xkb_keycode in iter(keymap):
        try:
            if xkb_keycode_to_ev_keycode(xkb_keycode) not in VALID_EV_KEYCODES:
                # Ignore keys that can't be sent or received by evdev.
                continue
            # Levels are different outputs from the same key with different modifiers pressed.
            level_count = keymap.num_levels_for_key(xkb_keycode, layout_index)

            for level in range(level_count):
                key_syms_for_level = keymap.key_get_syms_by_level(
                    xkb_keycode, layout_index, level
                )
                for key_sym in key_syms_for_level:
                    add_xkb_keysym_to_plover_keymap(
                        xkb_keycode,
                        key_sym,
                        level,
                        keymap,
                        plover_key_to_keycode,
                        layout_index,
                        modifier_index_to_xkb_keycode,
                    )

        except xkb.XKBInvalidKeycode:
            # Iter *should* return only valid, but still returns some invalid...
            pass

    # The "Linefeed" symbol (xkb symbol 0xff0a) has the key string "\n".
    # If Linefeed appears before the enter/return key when iterating over keys in the keymap (which is the case for qwerty), "\n" will be mapped to Linefeed rather than enter.
    # Ensures that "\n" is mapped to the enter/return key instead of Linefeed.
    if "return" in plover_key_to_keycode:
        plover_key_to_keycode["\n"] = plover_key_to_keycode["return"]

    return plover_key_to_keycode


def generate_plover_keymap_from_xkb_keymap(
    keymap: xkb.Keymap,
) -> dict[str, KeyCodeInfo]:
    """
    Wrapper around `generate_plover_keymap_from_xkb_keymap_and_modifiers` that computes the modifiers for you.
    """
    modifier_index_to_keycode = get_modifier_keycodes(keymap)
    return generate_plover_keymap_from_xkb_keymap_and_modifiers(
        keymap, modifier_index_to_keycode
    )


def add_xkb_keysym_to_plover_keymap(
    xkb_keycode: int,
    xkb_keysym: int,
    level: int,
    keymap: xkb.Keymap,
    plover_key_to_keycode: dict[str, KeyCodeInfo],
    layout_index: int,
    modifier_index_to_keycode: list[list[int]],
):
    keysym_name = xkb.keysym_get_name(xkb_keysym)
    keysym_string = xkb.keysym_to_string(xkb_keysym)

    key_modifiers = get_modifiers_for_key_sym(
        keymap, xkb_keycode, layout_index, level, modifier_index_to_keycode
    )

    if keysym_string is not None and keysym_string not in plover_key_to_keycode:
        # Because we iterate levels in order, the lowest level and thus simplest set of modifiers for each symbol is added first.
        # If multiple keys produce the same symbol, only add the first key in iteration order. Same for level_key_name and aliases below.
        plover_key_to_keycode[keysym_string] = KeyCodeInfo(
            xkb_keycode_to_ev_keycode(xkb_keycode), key_modifiers
        )

    for key_alias in XKB_KEY_NAME_TO_ALIASES.get(keysym_name, []):
        if key_alias not in plover_key_to_keycode:
            plover_key_to_keycode[key_alias] = KeyCodeInfo(
                xkb_keycode_to_ev_keycode(xkb_keycode), key_modifiers
            )

    if keysym_name.startswith("XF86"):
        plover_key_name = keysym_name[4:].lower()
        # Add alias with "xf86" for keys starting with "XF86" to be consistent with X11 Plover.
        if plover_key_name not in plover_key_to_keycode:
            plover_key_to_keycode[plover_key_name] = KeyCodeInfo(
                xkb_keycode_to_ev_keycode(xkb_keycode), key_modifiers
            )

    level_key_name_lower = keysym_name.lower()
    if level_key_name_lower not in plover_key_to_keycode:
        plover_key_to_keycode[level_key_name_lower] = KeyCodeInfo(
            xkb_keycode_to_ev_keycode(xkb_keycode), key_modifiers
        )


def get_modifiers_for_key_sym(
    keymap: xkb.Keymap,
    xkb_keycode: int,
    layout_index: int,
    level: int,
    modifier_index_to_keycode: list[list[int]],
) -> list[int]:
    """Get one set of modifiers that are pressed to obtain the given key and level.
    If multiple sets of modifiers produce the same key and level, an arbitrary one is returned."""
    modifier_masks_for_level = keymap.key_get_mods_for_level(
        xkb_keycode, layout_index, level
    )
    key_modifiers: list[int] = []
    # Identify sets of modifiers pressed to obtain the given key and level.
    # Each `mask` is a bitfield of modifiers pressed.
    for mask in modifier_masks_for_level:
        modifier_index = 0
        while mask > 0:
            if mask & 1:
                modifier_keycodes = modifier_index_to_keycode[modifier_index]
                if not modifier_keycodes:
                    # Invalid modifier index. Try the next mask.
                    key_modifiers.clear()
                    break
                key_modifiers.append(xkb_keycode_to_ev_keycode(modifier_keycodes[0]))
            mask >>= 1
            modifier_index += 1
        else:
            # Iterated through all modifiers in a mask and found a valid set.
            break
    return key_modifiers


_context = xkb.Context()

LAYOUTS = {
    "qwerty": generate_plover_keymap_from_xkb_keymap(
        _context.keymap_new_from_names(layout="us")
    ),
    "qwertz": generate_plover_keymap_from_xkb_keymap(
        _context.keymap_new_from_names(layout="de")
    ),
    "dvorak": generate_plover_keymap_from_xkb_keymap(
        _context.keymap_new_from_names(layout="us", variant="dvorak")
    ),
    "colemak": generate_plover_keymap_from_xkb_keymap(
        _context.keymap_new_from_names(layout="us", variant="colemak")
    ),
    "colemak-dh": generate_plover_keymap_from_xkb_keymap(
        _context.keymap_new_from_names(layout="us", variant="colemak_dh")
    ),
}

del _context

DEFAULT_LAYOUT = "qwerty"
assert DEFAULT_LAYOUT in LAYOUTS, "Default layout not in defined layouts"

# Linux EV keycode to Plover key name. Determines which keys can be handled and will be suppressed.
# Many key names are different from xkbcommon, so it's easier to define manually.
HANDLED_EV_KEYCODE_TO_KEY = {
    e.KEY_F1: "F1",
    e.KEY_F2: "F2",
    e.KEY_F3: "F3",
    e.KEY_F4: "F4",
    e.KEY_F5: "F5",
    e.KEY_F6: "F6",
    e.KEY_F7: "F7",
    e.KEY_F8: "F8",
    e.KEY_F9: "F9",
    e.KEY_F10: "F10",
    e.KEY_F11: "F11",
    e.KEY_F12: "F12",
    e.KEY_GRAVE: "`",
    e.KEY_0: "0",
    e.KEY_1: "1",
    e.KEY_2: "2",
    e.KEY_3: "3",
    e.KEY_4: "4",
    e.KEY_5: "5",
    e.KEY_6: "6",
    e.KEY_7: "7",
    e.KEY_8: "8",
    e.KEY_9: "9",
    e.KEY_MINUS: "-",
    e.KEY_EQUAL: "=",
    e.KEY_Q: "q",
    e.KEY_W: "w",
    e.KEY_E: "e",
    e.KEY_R: "r",
    e.KEY_T: "t",
    e.KEY_Y: "y",
    e.KEY_U: "u",
    e.KEY_I: "i",
    e.KEY_O: "o",
    e.KEY_P: "p",
    e.KEY_LEFTBRACE: "[",
    e.KEY_RIGHTBRACE: "]",
    e.KEY_BACKSLASH: "\\",
    e.KEY_A: "a",
    e.KEY_S: "s",
    e.KEY_D: "d",
    e.KEY_F: "f",
    e.KEY_G: "g",
    e.KEY_H: "h",
    e.KEY_J: "j",
    e.KEY_K: "k",
    e.KEY_L: "l",
    e.KEY_SEMICOLON: ";",
    e.KEY_APOSTROPHE: "'",
    e.KEY_Z: "z",
    e.KEY_X: "x",
    e.KEY_C: "c",
    e.KEY_V: "v",
    e.KEY_B: "b",
    e.KEY_N: "n",
    e.KEY_M: "m",
    e.KEY_COMMA: ",",
    e.KEY_DOT: ".",
    e.KEY_SLASH: "/",
    e.KEY_SPACE: "space",
    e.KEY_BACKSPACE: "BackSpace",
    e.KEY_DELETE: "Delete",
    e.KEY_DOWN: "Down",
    e.KEY_END: "End",
    e.KEY_ESC: "Escape",
    e.KEY_HOME: "Home",
    e.KEY_LEFT: "Left",
    e.KEY_PAGEDOWN: "Page_Down",
    e.KEY_PAGEUP: "Page_Up",
    e.KEY_ENTER: "Return",
    e.KEY_RIGHT: "Right",
    e.KEY_TAB: "Tab",
    e.KEY_UP: "Up",
}

# Make sure no keys missing. The last 3 are "\r\x0b\x0c" which don't need to be mapped.
assert all(c in LAYOUTS[DEFAULT_LAYOUT].keys() for c in string.printable[:-3])

if __name__ == "__main__":
    xkb_keymap = get_wayland_keymap(GET_WAYLAND_KEYMAP_TIMEOUT_SECONDS)
    plover_keymap = generate_plover_keymap_from_xkb_keymap(xkb_keymap)
    print("Plover keymap", plover_keymap)



from .blackbox import blackbox_test
from .dict import make_dict
from .output import CaptureOutput
from .parametrize import parametrize
from .steno import steno_to_stroke
from .steno_dictionary import dictionary_test

__all__ = [
    "blackbox_test",
    "make_dict",
    "CaptureOutput",
    "parametrize",
    "steno_to_stroke",
    "dictionary_test",
]



import array
import collections
import os
import selectors
import socket
import struct

from plover import log

WAYLAND_MESSAGE_HEADER_SIZE_BYTES = 8

# Wayland object IDs
DISPLAY_ID = 1
REGISTRY_ID = 2
SYNC_ID = 3
SEAT_ID = 4
KEYBOARD_ID = 5

# Wayland Opcodes
OPCODE_WL_DISPLAY_SYNC = 0
OPCODE_WL_DISPLAY_GET_REGISTRY = 1
OPCODE_WL_CALLBACK_DONE = 0
OPCODE_WL_REGISTRY_GLOBAL = 0
OPCODE_WL_REGISTRY_BIND = 0
OPCODE_WL_SEAT_CAPABILITIES = 0
OPCODE_WL_DISPLAY_ERROR = 0
OPCODE_WL_DISPLAY_DELETE_ID = 1
OPCODE_WL_KEYBOARD_KEYMAP = 0

WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1 = 1


def round_up_power_of_two(value: int, multiple: int):
    """Round `value` up to the nearest multiple of `multiple`.
    `multiple` must be positive and a power of 2"""
    assert (multiple > 0) and ((multiple & (multiple - 1)) == 0), (
        "Multiple must be positive and a power of two"
    )
    return (value + multiple - 1) & ~(multiple - 1)


class WaylandConnection:
    """Context manager for connecting to the Wayland server on the default socket path.

    Useful references:
    - https://wayland-book.com/
    - https://wayland.freedesktop.org/docs/html/ch04.html#sect-Protocol-Wire-Format
    - https://wayland.app/protocols/wayland
    """

    fd_queue: collections.deque[int]
    _wayland_socket: socket.socket
    _shutdown_pipe_read: int
    _shutdown_pipe_write: int
    _selector: selectors.BaseSelector

    def __init__(self):
        self.fd_queue = collections.deque()
        self._shutdown_pipe_read, self._shutdown_pipe_write = os.pipe()
        self._selector = selectors.DefaultSelector()

    def __enter__(self):
        # Find socket path following libwayland (https://wayland-book.com/protocol-design/wire-protocol.html#transports)
        xdg_runtime_dir = os.environ.get("XDG_RUNTIME_DIR", "")
        wayland_display = os.environ.get("WAYLAND_DISPLAY", "wayland-0")
        socket_path = os.path.join(xdg_runtime_dir, wayland_display)

        self._wayland_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._wayland_socket.connect(socket_path)

        self._selector.register(self._wayland_socket, selectors.EVENT_READ)
        self._selector.register(self._shutdown_pipe_read, selectors.EVENT_READ)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._selector.close()
        finally:
            self._wayland_socket.shutdown(socket.SHUT_RDWR)
            self._wayland_socket.close()
            os.close(self._shutdown_pipe_read)
            os.close(self._shutdown_pipe_write)

    def recv_message(self) -> tuple[int, int, int, bytearray]:
        """Receive an event from the Wayland server. Blocks until a complete message is received.

        Returns:
            A tuple of (object_id, length, opcode, event_data_bytes)

        length includes the message header.
        """
        # The only event with fds that we care about is wl_keyboard::keymap which only has one fd
        # In each message, we only need to receive at most one fd
        # TODO: Unless messages with more delay when we received the keymap event fds?
        MAX_FD_COUNT = 1
        event_header_bytes, fds = self._recv_fds_exact(
            WAYLAND_MESSAGE_HEADER_SIZE_BYTES, MAX_FD_COUNT
        )
        self.fd_queue.extend(fds)
        object_id, length_and_opcode = struct.unpack("=II", event_header_bytes)
        length = length_and_opcode >> 16
        assert length % 4 == 0, "Length of message must be a multiple of 4."
        opcode = length_and_opcode & 0xFFFF
        event_data_bytes, fds = self._recv_fds_exact(
            length - WAYLAND_MESSAGE_HEADER_SIZE_BYTES, MAX_FD_COUNT
        )
        self.fd_queue.extend(fds)
        return object_id, length, opcode, event_data_bytes

    def send_message(self, object_id: int, opcode: int, data: bytes | bytearray):
        """Send a request to the Wayland server.

        Args:
            object_id: The ID of the object to send the request to.
            opcode: The opcode of the request.
            data: The data to send with the request.
        """
        length = WAYLAND_MESSAGE_HEADER_SIZE_BYTES + len(data)
        # Wayland messages are streams of 32-bit (4 byte) values
        assert length % 4 == 0, "Length of message must be a multiple of 4."
        # The length field is a 16-bit unsigned integer (the upper 16 bits of the 32-bit value)
        assert length < 2**16, "Length of message must be less than 2^16."
        length_and_opcode = (length << 16) | opcode
        message = struct.pack("=II", object_id, length_and_opcode)
        self._wayland_socket.sendall(message)
        self._wayland_socket.sendall(data)

    def shutdown(self):
        """Signal the Wayland connection to close and for the event loop to exit."""
        os.write(self._shutdown_pipe_write, b"\x00")

    def _recv_fds_exact(self, length: int, fd_count: int):
        """Receive exactly `length` bytes from the Wayland server and up to `fd_count` file descriptors.

        Returns:
            A tuple of (data bytes received, fds received)
        Raises:
            InterruptedError: if the connection is shut down using `WaylandConnection.shutdown()`.
        """
        fds = array.array("i")
        buffer = bytearray(length)
        buffer_view = memoryview(buffer)

        if length < 0:
            raise ValueError("Length must be non-negative.")
        if fd_count < 0:
            raise ValueError("FD count must be non-negative.")

        while length > 0:
            for key, _ in self._selector.select():
                if key.fileobj == self._shutdown_pipe_read:
                    raise InterruptedError()
                # Based on Python3 socket.recvmsg docs (https://docs.python.org/3/library/socket.html#socket.socket.recvmsg)
                n, ancdata, flags, addr = self._wayland_socket.recvmsg_into(
                    [buffer_view], socket.CMSG_LEN(fd_count * fds.itemsize)
                )
                for cmsg_level, cmsg_type, cmsg_data in ancdata:
                    if (
                        cmsg_level == socket.SOL_SOCKET
                        and cmsg_type == socket.SCM_RIGHTS
                    ):
                        # Append data, ignoring any truncated integers at the end.
                        fds.frombytes(
                            cmsg_data[
                                : len(cmsg_data) - (len(cmsg_data) % fds.itemsize)
                            ]
                        )
                # Advance write position in buffer
                buffer_view = buffer_view[n:]
                length -= n

        fds = list(fds)

        return buffer, fds


def wayland_keymap_event_loop(connection: WaylandConnection) -> tuple[int, int]:
    """Get the keymap from the Wayland server.
    See https://wayland.app/protocols/wayland for the opcodes and arguments

    Returns a tuple of (keymap_fd, keymap_size) as returned by the Wayland server.
    """
    # wl_display::get_registry
    # display id: DISPLAY_ID
    # opcode: 1
    # new id for registry: REGISTRY_ID
    connection.send_message(
        DISPLAY_ID, OPCODE_WL_DISPLAY_GET_REGISTRY, struct.pack("=I", REGISTRY_ID)
    )

    # wl_display::sync
    # display id: DISPLAY_ID
    # opcode: 0
    # new_id for callback: SYNC_ID
    connection.send_message(
        DISPLAY_ID, OPCODE_WL_DISPLAY_SYNC, struct.pack("=I", SYNC_ID)
    )

    # Read all wl_display::get_registry events
    while True:
        object_id, length, opcode, event_data_bytes = connection.recv_message()
        if object_id == SYNC_ID:
            if opcode != OPCODE_WL_CALLBACK_DONE:
                raise RuntimeError(f"Expected wl_callback::done opcode 0, got {opcode}")
            break
        elif object_id == REGISTRY_ID and opcode == OPCODE_WL_REGISTRY_GLOBAL:
            # wl_registry::global
            name, interface_length = struct.unpack("=II", event_data_bytes[:8])
            # -1 to skip null terminator
            interface = event_data_bytes[8 : 8 + interface_length - 1].decode("utf-8")
            version_start_index = round_up_power_of_two(8 + interface_length, 4)
            version = struct.unpack(
                "=I", event_data_bytes[version_start_index : version_start_index + 4]
            )[0]
            log.debug(
                "Global: name=%d, interface=%s, version=%d", name, interface, version
            )

            if interface == "wl_seat":
                seat_name = event_data_bytes
                # Bind to seat using wl_registry::bind
                # opcode 0
                # the new_id arg follows custom serialization rules (interface name, version, id). See the representation of new_id in https://wayland.freedesktop.org/docs/html/ch04.html#sect-Protocol-Wire-Format
                # new id for seat: SEAT_ID
                data = seat_name + struct.pack("=I", SEAT_ID)
                connection.send_message(REGISTRY_ID, OPCODE_WL_REGISTRY_BIND, data)
        else:
            log.debug("Ignoring event for object %d, opcode %d", object_id, opcode)

    # Read wl_seat events
    has_keyboard = False
    while True:
        object_id, length, opcode, event_data_bytes = connection.recv_message()
        if object_id == SEAT_ID and opcode == OPCODE_WL_SEAT_CAPABILITIES:
            # wl_seat::capabilities
            if length != WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 4:
                raise RuntimeError(
                    f"Expected wl_seat::capabilities message to be {WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 4} bytes, got {length}"
                )

            capabilities = struct.unpack("=I", event_data_bytes[:4])[0]
            log.debug("wl_seat capabilities: %s", capabilities)
            has_keyboard = capabilities & 2
            break
        elif object_id == DISPLAY_ID and opcode == OPCODE_WL_DISPLAY_ERROR:
            # wl_display::error
            raise RuntimeError(f"Wayland error: {repr(event_data_bytes)}")
        elif object_id == DISPLAY_ID and opcode == OPCODE_WL_DISPLAY_DELETE_ID:
            # wl_display::delete_id
            if length != WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 4:
                raise RuntimeError(
                    f"Expected wl_display::delete_id message to be {WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 4} bytes, got {length}"
                )
            id_num = struct.unpack("=I", event_data_bytes)[0]
            if id_num == SEAT_ID:
                raise RuntimeError("wl_seat was destroyed unexpectedly")
            elif id_num == KEYBOARD_ID:
                raise RuntimeError("wl_keyboard was destroyed unexpectedly")
        else:
            log.debug("Ignoring event for object %d, opcode %d", object_id, opcode)

    if not has_keyboard:
        raise RuntimeError("Wayland seat has no keyboard")

    # wl_seat::get_keyboard
    connection.send_message(SEAT_ID, 1, struct.pack("=I", KEYBOARD_ID))

    # Wait for and process wl_keyboard::keymap
    while True:
        object_id, length, opcode, event_data_bytes = connection.recv_message()
        if object_id == KEYBOARD_ID and opcode == OPCODE_WL_KEYBOARD_KEYMAP:
            # wl_keyboard::keymap
            if length != WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 8:
                raise RuntimeError(
                    f"Expected wl_keyboard::keymap message to be {WAYLAND_MESSAGE_HEADER_SIZE_BYTES + 8} bytes, got {length}"
                )

            keymap_format, keymap_size = struct.unpack("=II", event_data_bytes)
            if keymap_format != WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1:
                raise RuntimeError(f"Unsupported keymap format: {keymap_format}")

            try:
                fd = connection.fd_queue.popleft()
            except IndexError:
                raise RuntimeError("No keymap fd received")

            return fd, keymap_size
        elif object_id == DISPLAY_ID and opcode == OPCODE_WL_DISPLAY_ERROR:
            # wl_display::error
            raise RuntimeError(f"Wayland error: {repr(event_data_bytes)}")
        else:
            log.debug("Ignoring event for object %d, opcode %d", object_id, opcode)



"""This module provides a skeleton for implementing keyboard input to grab steno
keystrokes. This is typically platform-specific, so actual implementations
are contained in platform-specific :mod:`oslayer <plover.oslayer>` modules.
"""

from __future__ import annotations

from collections.abc import Sequence


class Capture:
    """Encapsulates logic for capturing keyboard input. An instance of this is
    used internally by Plover's built-in keyboard plugin.

    Define the :meth:`key_down` and :meth:`key_up` methods below to implement
    custom behavior that gets executed when a key is pressed or released.
    """

    def start(self) -> None:
        """Start collecting keyboard input."""
        raise NotImplementedError()

    def cancel(self) -> None:
        """Stop collecting keyboard input."""
        raise NotImplementedError()

    def suppress(self, suppressed_keys: Sequence[str] = ()) -> None:
        """Suppresses the specified keys, preventing them from returning any
        output through regular typing. This allows us to intercept keyboard
        events when using keyboard input.
        """
        raise NotImplementedError()

    # Callbacks for keyboard press/release events.
    def key_down(self, key: str) -> None:
        """Notifies Plover that a key was pressed down."""
        return None

    def key_up(self, key: str) -> None:
        """Notifies Plover that a key was released."""
        return None



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""A module to handle logging."""

import os
import sys
import logging
import traceback

from logging.handlers import RotatingFileHandler
from logging import INFO, WARNING

from plover.oslayer.config import CONFIG_DIR


LOG_FORMAT = "%(asctime)s [%(threadName)s] %(levelname)s: %(message)s"
LOG_FILENAME = os.path.realpath(os.path.join(CONFIG_DIR, "plover.log"))
LOG_MAX_BYTES = 10000000
LOG_COUNT = 9

STROKE_LOG_FORMAT = "%(asctime)s %(message)s"


class NoExceptionTracebackFormatter(logging.Formatter):
    """Custom formatter for formatting exceptions without traceback."""

    def format(self, record):
        # Calls to formatException are cached.
        # (see http://bugs.python.org/issue1295)
        orig_exc_text = record.exc_text
        record.exc_text = None
        try:
            return super().format(record)
        finally:
            record.exc_text = orig_exc_text

    def formatException(self, exc_info):
        etype, evalue, tb = exc_info
        lines = traceback.format_exception_only(etype, evalue)
        return "".join(lines)


class FileHandler(RotatingFileHandler):
    def __init__(self, filename=LOG_FILENAME, format=LOG_FORMAT):
        super().__init__(
            filename, maxBytes=LOG_MAX_BYTES, backupCount=LOG_COUNT, encoding="utf-8"
        )
        self.setFormatter(logging.Formatter(format))


class PrintHandler(logging.StreamHandler):
    """Handler using L{print_} to output messages."""

    def __init__(self, format=LOG_FORMAT):
        super().__init__(sys.stderr)
        self.setFormatter(logging.Formatter(format))


class Logger:
    def __init__(self):
        self._print_handler = PrintHandler()
        self._print_handler.setLevel(WARNING)
        self._file_handler = None
        self._platform_handler = None
        self._logger = logging.getLogger("plover")
        self._logger.addHandler(self._print_handler)
        self._logger.setLevel(INFO)
        self._stroke_filename = None
        self._stroke_logger = logging.getLogger("plover-strokes")
        self._stroke_logger.setLevel(INFO)
        self._stroke_handler = None
        self._log_strokes = False
        self._log_translations = False

    def has_platform_handler(self):
        return self._platform_handler is not None

    def setup_platform_handler(self):
        if self.has_platform_handler():
            return
        NotificationHandler = None
        try:
            from plover.oslayer.log import NotificationHandler
        except Exception:
            self.info("could not import platform gui log", exc_info=True)
            return
        try:
            handler = NotificationHandler()
            self.addHandler(handler)
        except Exception:
            self.info("could not initialize platform gui log", exc_info=True)
            return
        self._platform_handler = handler

    def set_level(self, level):
        self._print_handler.setLevel(level)
        if self._file_handler is not None:
            self._file_handler.setLevel(level)
        self.setLevel(level)

    def setup_logfile(self):
        assert self._file_handler is None
        self._file_handler = FileHandler()
        self._file_handler.setLevel(self.level)
        self._logger.addHandler(self._file_handler)

    def _setup_stroke_logging(self):
        is_logging = self._stroke_handler is not None
        must_log = (
            self._log_strokes or self._log_translations
        ) and self._stroke_filename is not None
        if must_log:
            if is_logging:
                start_logging = stop_logging = (
                    self._stroke_filename != self._stroke_handler.baseFilename
                )
            else:
                stop_logging = False
                start_logging = True
        else:
            stop_logging = is_logging
            start_logging = False
        if stop_logging:
            self._stroke_logger.removeHandler(self._stroke_handler)
            self._stroke_handler.close()
            self._stroke_handler = None
        if start_logging:
            self._stroke_handler = FileHandler(
                filename=self._stroke_filename, format=STROKE_LOG_FORMAT
            )
            self._stroke_logger.addHandler(self._stroke_handler)

    def set_stroke_filename(self, filename=None):
        if filename is not None:
            filename = os.path.realpath(filename)
        if self._stroke_filename == filename:
            return
        assert filename != LOG_FILENAME
        self.info("set_stroke_filename(%s)", filename)
        self._stroke_filename = filename
        self._setup_stroke_logging()

    def enable_stroke_logging(self, enable):
        if self._log_strokes == enable:
            return
        self.info("enable_stroke_logging(%s)", enable)
        self._log_strokes = enable
        self._setup_stroke_logging()

    def enable_translation_logging(self, enable):
        if self._log_translations == enable:
            return
        self.info("enable_translation_logging(%s)", enable)
        self._log_translations = enable
        self._setup_stroke_logging()

    def log_stroke(self, stroke):
        if not self._log_strokes or self._stroke_handler is None:
            return
        self._stroke_logger.info("%s", stroke)

    def log_translation(self, undo, do, prev):
        if not self._log_translations or self._stroke_handler is None:
            return
        # TODO: Figure out what to actually log here.
        for u in undo:
            self._stroke_logger.info("*%s", u)
        for d in do:
            self._stroke_logger.info(d)

    # Delegate calls to _logger.
    def __getattr__(self, name):
        return getattr(self._logger, name)


# Set up default logger.
__logger = Logger()

# The following functions direct all input to __logger.
debug = __logger.debug
info = __logger.info
warning = __logger.warning
error = __logger.error
set_level = __logger.set_level
add_handler = __logger.addHandler
remove_handler = __logger.removeHandler
has_platform_handler = __logger.has_platform_handler
setup_platform_handler = __logger.setup_platform_handler
# Strokes/translation logging.
set_stroke_filename = __logger.set_stroke_filename
stroke = __logger.log_stroke
translation = __logger.log_translation
enable_stroke_logging = __logger.enable_stroke_logging
enable_translation_logging = __logger.enable_translation_logging
# Logfile support.
setup_logfile = __logger.setup_logfile



from typing import Optional
from plover import log

from PySide6.QtCore import Qt
from PySide6.QtGui import QGuiApplication

MODE_SYSTEM = "system"
MODE_LIGHT = "light"
MODE_DARK = "dark"


def _set_mode(mode: Optional[str]) -> None:
    """Apply the requested mode to the current Qt application.

    This function tweaks Qt's idea of the platform color scheme via
    QStyleHints.setColorScheme / unsetColorScheme so that other code can
    query styleHints().colorScheme() and get the effective setting.

    - "system": follow the operating system setting again
    - "light":  force a light color scheme
    - "dark":   force a dark color scheme
    - None:     no-op
    """

    if mode is None:
        # Explicitly do nothing when no mode was provided.
        return

    app = QGuiApplication.instance()
    if not isinstance(app, QGuiApplication):
        # No GUI application instance (or unexpected type); nothing to do.
        return

    # Ensure the style is initialized.
    hints = app.styleHints()

    if mode == MODE_SYSTEM:
        # Remove any explicit override so Qt follows the OS again.
        hints.unsetColorScheme()
    elif mode == MODE_LIGHT:
        hints.setColorScheme(Qt.ColorScheme.Light)
    elif mode == MODE_DARK:
        hints.setColorScheme(Qt.ColorScheme.Dark)
    else:
        log.warning(f"invalid mode config value: {mode}")


def update(config):
    _set_mode(config.get("appearance_mode"))



"""The steno engine is the core of Plover; it handles communication between the
machine and the translation and formatting subsystems, and manages configuration
and dictionaries.
"""

from collections import namedtuple, OrderedDict
from functools import wraps
from queue import Queue
import functools
import os
import shutil
import threading
from typing import Any, Callable, Dict, List, Optional, Tuple

from plover import log, system
from plover.dictionary.loading_manager import DictionaryLoadingManager
from plover.formatting import Formatter
from plover.misc import shorten_path
from plover.registry import registry
from plover.resource import ASSET_SCHEME, resource_filename
from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.suggestions import Suggestions
from plover.translation import Translator


StartingStrokeState = namedtuple(
    "StartingStrokeState", "attach capitalize space_char", defaults=(False, False, " ")
)

StartingStrokeState.__doc__ = """An object representing the starting state of the formatter before any
strokes are input.

Attributes:
    attach (bool): Whether to delete the space before the translation when the
        initial stroke is translated.
    capitalize (bool): Whether to capitalize the translation when the initial
        stroke is translated.
"""

MachineParams = namedtuple("MachineParams", "type options keymap")

MachineParams.__doc__ = """An object representing the current state of the machine.

Attributes:
    type (str): The name of the machine. This is the same as the name of the plugin
        that provides the machine's functionality. ``Keyboard`` by default.
    options (Dict[str, Any]): A dictionary of machine specific options. See
        :mod:`plover.config` for more information.
    keymap (plover.machine.keymap.Keymap): A
        :class:`Keymap<plover.machine.keymap.Keymap>` mapping the current
        system to this machine.
"""


class ErroredDictionary(StenoDictionary):
    """A placeholder class for a dictionary that failed to load.

    This is a subclass of :class:`StenoDictionary<plover.steno_dictionary.StenoDictionary>`.

    Attributes:
        path (str): The path to the dictionary file.
        exception (Any): The exception that caused the dictionary loading to fail.
    """

    def __init__(self, path: str, exception: Any):
        super().__init__()
        self.enabled = False
        self.readonly = True
        self.path = path
        self.exception = exception

    def __eq__(self, other):
        if not isinstance(other, ErroredDictionary):
            return False
        return (self.path, self.exception) == (other.path, other.exception)


def copy_default_dictionaries(dictionaries_files):
    """Recreate default dictionaries.

    Each default dictionary is recreated if it's
    in use by the current config and missing.
    """

    for dictionary in dictionaries_files:
        # Ignore assets.
        if dictionary.startswith(ASSET_SCHEME):
            continue
        # Nothing to do if dictionary file already exists.
        if os.path.exists(dictionary):
            continue
        # Check it's actually a default dictionary.
        basename = os.path.basename(dictionary)
        if basename not in system.DEFAULT_DICTIONARIES:
            continue
        default_dictionary = os.path.join(system.DICTIONARIES_ROOT, basename)
        log.info("recreating %s from %s", dictionary, default_dictionary)
        shutil.copyfile(resource_filename(default_dictionary), dictionary)


def with_lock(func):
    # To keep __doc__/__name__ attributes of the initial function.
    @wraps(func)
    def _with_lock(self, *args, **kwargs):
        with self:
            return func(self, *args, **kwargs)

    return _with_lock


class StenoEngine:
    """
    Attributes:
        config (Dict[str, Any]): A dictionary containing configuration options.
        controller (plover.oslayer.controller.Controller): An instance of
            :class:`Controller<plover.oslayer.controller.Controller>` for managing
            commands sent to this Plover instance. This is provided during startup.
        keyboard_emulation (plover.oslayer.keyboardcontrol.KeyboardEmulation): An
            instance of
            :class:`KeyboardEmulation<plover.oslayer.keyboardcontrol.KeyboardEmulation>`
            provided during startup.
        HOOKS (List[str]): A list of all the possible engine hooks. See
            :ref:`engine-hooks` below for a list of valid hooks.
        machine_state (str): The connection state of the current machine. One of
            ``stopped``, ``initializing``, ``connected`` or ``disconnected``.
        output (bool): ``True`` if steno output is enabled, ``False`` otherwise.
        _config (plover.config.Config): A :class:`Config<plover.config.Config>` object
            containing the engine's configuration.
        translator_state (plover.translation._State): A
            :class:`_State<plover.translation._State>` object containing the current
            state of the translator.
        starting_stroke_state (StartingStrokeState): A :class:`StartingStrokeState`
            representing the initial state of the formatter.
        dictionaries (plover.steno_dictionary.StenoDictionaryCollection): A
            :class:`StenoDictionaryCollection<plover.steno_dictionary.StenoDictionaryCollection>`
            of all the dictionaries Plover has loaded for the current system. This
            includes disabled dictionaries and dictionaries that failed to load.

    """

    HOOKS: List[str] = """
    stroked
    translated
    machine_state_changed
    output_changed
    config_changed
    dictionaries_loaded
    dictionary_state_changed
    send_string
    send_backspaces
    send_key_combination
    add_translation
    focus
    configure
    lookup
    suggestions
    quit
    """.split()

    def __init__(self, config: Any, controller: Any, keyboard_emulation: Any):
        self._config = config
        self._controller = controller
        self._is_running = False
        self._queue = Queue()
        self._lock = threading.RLock()
        self._machine = None
        self._machine_state = None
        self._machine_params = MachineParams(None, None, None)
        self._formatter = Formatter()
        self._formatter.set_output(
            Formatter.output_type(
                self._send_backspaces,
                self._send_string,
                self._send_key_combination,
                self._send_engine_command,
            )
        )
        self._formatter.add_listener(self._on_translated)
        self._translator = Translator()
        self._translator.add_listener(log.translation)
        self._translator.add_listener(self._formatter.format)
        self._dictionaries = self._translator.get_dictionary()  # type: StenoDictionaryCollection
        self._dictionaries_manager = DictionaryLoadingManager(
            functools.partial(self._trigger_hook, "dictionary_state_changed")
        )
        self._running_state = self._translator.get_state()
        self._translator.clear_state()
        self._keyboard_emulation = keyboard_emulation
        self._hooks = {hook: [] for hook in self.HOOKS}
        self._running_extensions = {}

    def __enter__(self):
        self._lock.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.__exit__(exc_type, exc_value, traceback)

    def _in_engine_thread(self) -> bool:
        """Returns whether we are currently in the same thread that the engine
        is running on.

        This is useful because event listeners for machines and others are run
        on separate threads, and we want to be able to run engine events on the
        same thread as the main engine.
        """
        raise NotImplementedError()

    def _same_thread_hook(self, func, *args, **kwargs):
        if self._in_engine_thread():
            func(*args, **kwargs)
        else:
            self._queue.put((func, args, kwargs))

    def run(self) -> None:
        """Starts the steno engine, translating any strokes that are input."""
        while True:
            func, args, kwargs = self._queue.get()
            try:
                with self._lock:
                    if func(*args, **kwargs):
                        break
            except Exception:
                log.error("engine %s failed", func.__name__[1:], exc_info=True)

    def _on_control_message(self, msg):
        if msg[0] == "command":
            self._same_thread_hook(self._execute_engine_command, *msg[1:], force=True)
        else:
            log.error("ignoring invalid control message: %r", msg)

    def _stop(self):
        self._controller.stop()
        self._stop_extensions(self._running_extensions.keys())
        if self._machine is not None:
            self._machine.stop_capture()
            self._machine = None

    def _start(self):
        self._set_output(self._config["auto_start"])
        self._update(full=True)
        self._controller.start(self._on_control_message)

    def _set_dictionaries(self, dictionaries):
        def dictionaries_changed(l1, l2):
            if len(l1) != len(l2):
                return True
            for d1, d2 in zip(l1, l2):
                if d1 is not d2:
                    return True
            return False

        if not dictionaries_changed(dictionaries, self._dictionaries.dicts):
            # No change.
            return
        self._dictionaries.set_dicts(dictionaries)
        self._trigger_hook(
            "dictionaries_loaded", StenoDictionaryCollection(dictionaries)
        )

    def _update(self, config_update=None, full=False, reset_machine=False):
        original_config = self._config.as_dict()
        # Update configuration.
        if config_update is not None:
            self._config.update(**config_update)
            config = self._config.as_dict()
        else:
            config = original_config
        # Create configuration update.
        if full:
            config_update = config
        else:
            config_update = {
                option: value
                for option, value in config.items()
                if value != original_config[option]
            }
            # Save config if anything changed.
            if config_update:
                self._config.save()
        # Update logging.
        log.set_stroke_filename(config["log_file_name"])
        log.enable_stroke_logging(config["enable_stroke_logging"])
        log.enable_translation_logging(config["enable_translation_logging"])
        # Update output.
        self._formatter.set_space_placement(config["space_placement"])
        self._formatter.start_attached = config["start_attached"]
        self._formatter.start_capitalized = config["start_capitalized"]
        self._translator.set_min_undo_length(config["undo_levels"])
        self._keyboard_emulation.set_key_press_delay(config["time_between_key_presses"])
        # This only applies to UInput, because it emulates a physical keyboard and follows the layout set in software. Because there is no standard of defining it, the user has to do so manually if not using a QWERTY keyboard.
        if hasattr(self._keyboard_emulation, "_update_layout"):
            self._keyboard_emulation._update_layout(config["keyboard_layout"])
        # Update system.
        system_name = config["system_name"]
        if system.NAME != system_name:
            log.info("loading system: %s", system_name)
            system.setup(system_name)
        # Update machine.
        update_keymap = False
        start_machine = False
        machine_params = MachineParams(
            config["machine_type"],
            config["machine_specific_options"],
            config["system_keymap"],
        )
        # Do not reset if only the keymap changed.
        if (
            self._machine_params is None
            or self._machine_params.type != machine_params.type
            or self._machine_params.options != machine_params.options
        ):
            reset_machine = True
        if reset_machine:
            if self._machine is not None:
                self._machine.stop_capture()
                self._machine = None
            machine_class = registry.get_plugin("machine", machine_params.type).obj
            log.info("setting machine: %s", machine_params.type)
            self._machine = machine_class(machine_params.options)
            self._machine.set_suppression(self._is_running)
            self._machine.add_state_callback(self._machine_state_callback)
            self._machine.add_stroke_callback(self._machine_stroke_callback)
            self._machine_params = machine_params
            update_keymap = True
            start_machine = True
        elif self._machine is not None:
            update_keymap = "system_keymap" in config_update
        if update_keymap:
            machine_keymap = config["system_keymap"]
            if machine_keymap is not None:
                self._machine.set_keymap(machine_keymap)
        if start_machine:
            self._machine.start_capture()
        # Update running extensions.
        enabled_extensions = config["enabled_extensions"]
        running_extensions = set(self._running_extensions)
        self._stop_extensions(running_extensions - enabled_extensions)
        self._start_extensions(enabled_extensions - running_extensions)
        # Trigger `config_changed` hook.
        if config_update:
            self._trigger_hook("config_changed", config_update)
        # Update dictionaries.
        config_dictionaries = OrderedDict((d.path, d) for d in config["dictionaries"])
        copy_default_dictionaries(config_dictionaries.keys())
        # Start by unloading outdated dictionaries.
        self._dictionaries_manager.unload_outdated()
        self._set_dictionaries(
            [
                d
                for d in self._dictionaries.dicts
                if d.path in config_dictionaries
                and d.path in self._dictionaries_manager
            ]
        )
        # And then (re)load all dictionaries.
        dictionaries = []
        for d in self._dictionaries_manager.load(config_dictionaries.keys()):
            if isinstance(d, ErroredDictionary):
                # Only show an error if it's new.
                if d != self._dictionaries.get(d.path):
                    log.error(
                        "loading dictionary `%s` failed: %s",
                        shorten_path(d.path),
                        str(d.exception),
                    )
            d.enabled = config_dictionaries[d.path].enabled
            dictionaries.append(d)
        self._set_dictionaries(dictionaries)

    def _start_extensions(self, extension_list):
        for extension_name in extension_list:
            log.info("starting `%s` extension", extension_name)
            try:
                extension = registry.get_plugin("extension", extension_name).obj(self)
            except KeyError:
                # Plugin not installed, skip.
                continue
            try:
                extension.start()
            except Exception:
                log.error(
                    "initializing extension `%s` failed", extension_name, exc_info=True
                )
            else:
                self._running_extensions[extension_name] = extension

    def _stop_extensions(self, extension_list):
        for extension_name in list(extension_list):
            log.info("stopping `%s` extension", extension_name)
            extension = self._running_extensions.pop(extension_name)
            extension.stop()
            del extension

    def _quit(self, code):
        self._stop()
        self.code = code
        self._trigger_hook("quit")
        return True

    def _toggle_output(self):
        self._set_output(not self._is_running)

    def _set_output(self, enabled):
        if enabled == self._is_running:
            return
        self._is_running = enabled
        if enabled:
            self._translator.set_state(self._running_state)
        else:
            self._translator.clear_state()
        if self._machine is not None:
            self._machine.set_suppression(enabled)
        self._trigger_hook("output_changed", enabled)

    def _machine_state_callback(self, machine_state):
        self._same_thread_hook(self._on_machine_state_changed, machine_state)

    def _machine_stroke_callback(self, steno_keys):
        self._same_thread_hook(self._on_stroked, steno_keys)

    @with_lock
    def _on_machine_state_changed(self, machine_state):
        assert machine_state is not None
        self._machine_state = machine_state
        self._trigger_hook(
            "machine_state_changed", self._machine_params.type, machine_state
        )

    def _consume_engine_command(self, command, force=False):
        # The first commands can be used whether plover has output enabled or not.
        command_name, *command_args = command.split(":", 1)
        command_name = command_name.lower()
        if command_name == "resume":
            self._set_output(True)
            return True
        elif command_name == "toggle":
            self._toggle_output()
            return True
        elif command_name == "quit":
            self.quit()
            return True
        if not force and not self._is_running:
            return False
        # These commands can only be run when plover has output enabled.
        if command_name == "suspend":
            self._set_output(False)
        elif command_name == "configure":
            self._trigger_hook("configure")
        elif command_name == "focus":
            self._trigger_hook("focus")
        elif command_name == "add_translation":
            self._trigger_hook("add_translation")
        elif command_name == "lookup":
            self._trigger_hook("lookup")
        elif command_name == "suggestions":
            self._trigger_hook("suggestions")
        else:
            command_fn = registry.get_plugin("command", command_name).obj
            command_fn(self, command_args[0] if command_args else "")
        return False

    def _execute_engine_command(self, command, force=False):
        self._consume_engine_command(command, force=force)
        return False

    def _on_stroked(self, steno_keys):
        stroke = Stroke(steno_keys)
        log.stroke(stroke)
        self._translator.translate(stroke)
        self._trigger_hook("stroked", stroke)

    def _on_translated(self, old, new):
        if not self._is_running:
            return
        self._trigger_hook("translated", old, new)

    def _send_backspaces(self, b):
        if not self._is_running:
            return
        self._keyboard_emulation.send_backspaces(b)
        self._trigger_hook("send_backspaces", b)

    def _send_string(self, s):
        if not self._is_running:
            return
        self._keyboard_emulation.send_string(s)
        self._trigger_hook("send_string", s)

    def _send_key_combination(self, c):
        if not self._is_running:
            return
        self._keyboard_emulation.send_key_combination(c)
        self._trigger_hook("send_key_combination", c)

    def _send_engine_command(self, command: str) -> None:
        """Runs the specified Plover command, which can be either a built-in
        command like ``set_config`` or one from an external plugin.

        ``command`` is a string containing the command and its argument (if any),
        separated by a colon. For example, ``lookup`` sends the ``lookup`` command
        (the same as stroking ``{PLOVER:LOOKUP}``), and ``run_shell:foo`` sends the
        ``run_shell`` command with the argument ``foo``.
        """
        suppress = not self._is_running
        suppress &= self._consume_engine_command(command)
        if suppress:
            self._machine.suppress_last_stroke(self._keyboard_emulation.send_backspaces)

    def toggle_output(self):
        """Toggles steno mode.

        See :attr:`output` to get the current state, or
        :meth:`set_output` to set the state to a specific value.
        """
        self._same_thread_hook(self._toggle_output)

    def set_output(self, enabled: bool) -> None:
        """Enables or disables steno mode.

        Set ``enabled`` to ``True`` to enable steno mode, or ``False`` to disable it.
        """
        self._same_thread_hook(self._set_output, enabled)

    @property
    @with_lock
    def machine_state(self) -> Optional[str]:
        """The connection state of the current machine.

        One of ``stopped``, ``initializing``, ``connected`` or ``disconnected``.
        """
        return self._machine_state

    @property
    @with_lock
    def output(self) -> bool:
        """``True`` if steno output is enabled, ``False`` otherwise."""
        return self._is_running

    @output.setter
    def output(self, enabled):
        self._same_thread_hook(self._set_output, enabled)

    @property
    @with_lock
    def config(self) -> Dict[str, Any]:
        """A dictionary containing configuration options."""
        return self._config.as_dict()

    @config.setter
    def config(self, update):
        self._same_thread_hook(self._update, config_update=update)

    @with_lock
    def __getitem__(self, setting: str) -> Any:
        """Returns the value of the configuration property ``setting``."""
        return self._config[setting]

    def __setitem__(self, setting: str, value: Any) -> None:
        """Sets the configuration property ``setting`` to ``value``."""
        self.config = {setting: value}

    def reset_machine(self) -> None:
        """Resets the machine state and Plover's connection with the machine, if
        necessary, and loads all the configuration and dictionaries.
        """
        self._same_thread_hook(self._update, reset_machine=True)

    def load_config(self) -> bool:
        """Loads the Plover configuration file and returns ``True`` if it was
        loaded successfully, ``False`` if not.
        """
        try:
            self._config.load()
        except Exception:
            log.error(
                "loading configuration failed, resetting to default", exc_info=True
            )
            self._config.clear()
            return False
        return True

    def start(self) -> None:
        """Starts the steno engine."""
        self._same_thread_hook(self._start)

    def quit(self, code: int = 0) -> None:
        """Quits the steno engine, ensuring that all pending tasks are completed
        before exiting.
        """
        # We need to go through the queue, even when already called
        # from the engine thread so _quit's return code does break
        # the thread out of its main loop.
        self._queue.put((self._quit, (code,), {}))

    def restart(self) -> None:
        """Quits and restarts the steno engine, ensuring that all pending tasks
        are completed.
        """
        self.quit(-1)

    def join(self) -> int:
        """Joins any sub-threads if necessary and returns an exit code."""
        return self.code

    @with_lock
    def lookup(self, translation: Tuple[str, ...]) -> str:
        """Returns the first translation for the steno outline ``translation`` using
        all the filters.
        """
        return self._dictionaries.lookup(translation)

    @with_lock
    def raw_lookup(self, translation: Tuple[str, ...]) -> str:
        """Like :meth:`lookup`, but without any of the filters."""
        return self._dictionaries.raw_lookup(translation)

    @with_lock
    def lookup_from_all(self, translation: Tuple[str, ...]):
        """Returns all translations for the steno outline ``translation`` using
        all the filters.
        """
        return self._dictionaries.lookup_from_all(translation)

    @with_lock
    def raw_lookup_from_all(self, translation: Tuple[str, ...]):
        """Like :meth:`lookup_from_all`, but without any of the filters."""
        return self._dictionaries.raw_lookup_from_all(translation)

    @with_lock
    def reverse_lookup(self, translation: str):
        """Returns the list of steno outlines that translate to ``translation``."""
        return self._dictionaries.reverse_lookup(translation)

    @with_lock
    def casereverse_lookup(self, translation: str):
        """Like :meth:`reverse_lookup`, but performs a case-insensitive lookup."""
        return self._dictionaries.casereverse_lookup(translation)

    @with_lock
    def add_dictionary_filter(
        self, dictionary_filter: Callable[[Tuple[str, ...], str], bool]
    ) -> None:
        """Adds ``dictionary_filter`` to the list of dictionary filters.

        See :attr:`StenoDictionaryCollection.filters<plover.steno_dictionary.StenoDictionaryCollection.filters>`
        for more information.
        """
        self._dictionaries.add_filter(dictionary_filter)

    @with_lock
    def remove_dictionary_filter(
        self, dictionary_filter: Callable[[Tuple[str, ...], str], bool]
    ) -> None:
        """Removes ``dictionary_filter`` from the list of dictionary filters."""
        self._dictionaries.remove_filter(dictionary_filter)

    @with_lock
    def get_suggestions(self, translation: str):
        """Returns a list of suggestions for the specified ``translation``."""
        return Suggestions(self._dictionaries).find(translation)

    @property
    @with_lock
    def translator_state(self):
        """A :class:`_State<plover.translation._State>` object containing the
        current state of the translator.
        """
        return self._translator.get_state()

    @translator_state.setter
    @with_lock
    def translator_state(self, state):
        self._translator.set_state(state)

    @with_lock
    def clear_translator_state(self, undo: bool = False) -> None:
        """Resets the translator to an empty state, as if Plover had just started up,
        clearing the entire translation stack.

        If ``undo`` is ``True``, this also reverts all previous translations on the
        stack (which could include a lot of backspaces).
        """
        if undo:
            state = self._translator.get_state()
            if state.translations:
                self._formatter.format(state.translations, (), None)
        self._translator.clear_state()

    @property
    @with_lock
    def starting_stroke_state(self) -> StartingStrokeState:
        """A :class:`StartingStrokeState` representing the initial state of the
        formatter.
        """
        return StartingStrokeState(
            self._formatter.start_attached,
            self._formatter.start_capitalized,
            self._formatter.space_char,
        )

    @starting_stroke_state.setter
    @with_lock
    def starting_stroke_state(self, state):
        self._formatter.start_attached = state.attach
        self._formatter.start_capitalized = state.capitalize
        self._formatter.space_char = state.space_char

    @with_lock
    def add_translation(
        self,
        strokes: Tuple[str, ...],
        translation: str,
        dictionary_path: Optional[str] = None,
    ) -> None:
        """Adds a steno entry mapping the steno outline ``strokes`` to
        ``translation`` in the dictionary at ``dictionary_path``, if specified,
        or the first writable dictionary.
        """
        if dictionary_path is None:
            dictionary_path = self._dictionaries.first_writable().path
        self._dictionaries.set(strokes, translation, path=dictionary_path)
        self._dictionaries.save(path_list=(dictionary_path,))

    @property
    @with_lock
    def dictionaries(self):
        """A
        :class:`StenoDictionaryCollection<plover.steno_dictionary.StenoDictionaryCollection>`
        of all the dictionaries Plover has loaded for the current system.

        This includes disabled dictionaries and dictionaries that failed to load.
        """
        return self._dictionaries

    # Hooks.

    def _trigger_hook(self, hook, *args, **kwargs):
        for callback in self._hooks[hook]:
            try:
                callback(*args, **kwargs)
            except Exception:
                log.error("hook %r callback %r failed", hook, callback, exc_info=True)

    @with_lock
    def hook_connect(self, hook: str, callback: Callable[..., Any]) -> None:
        """Adds ``callback`` to the list of handlers that are called when the ``hook``
        hook gets triggered. Raises a ``KeyError`` if ``hook`` is not in
        :data:`HOOKS`.

        The expected signature of the callback is depends on the hook; see
        :ref:`engine-hooks` for more information.
        """
        self._hooks[hook].append(callback)

    @with_lock
    def hook_disconnect(self, hook: str, callback: Callable[..., Any]) -> None:
        """Removes ``callback`` from the list of handlers that are called when
        the ``hook`` hook is triggered. Raises a ``KeyError`` if ``hook`` is not in
        :data:`HOOKS`, and a ``ValueError`` if ``callback`` was never added as
        a handler in the first place.
        """
        self._hooks[hook].remove(callback)
